/*
    Bantumi
    Copyright 2005, 2006 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsj
    martin@martin.st
*/

#include <coemain.h>
#include "bantumiappview.h"
#include <flogger.h>
#include "bantumi.h"
#include "bantumigl.h"
#include "glfont.h"
#include <hal.h>
#include <eikenv.h>
#include <GLES/gl.h>
#include <aknutils.h>
#include "bantumiappui.h"

#include <time.h>
#include <stdlib.h>

GLFont *makeSymbianGLFont(CFbsBitGc *gc, int height);

void debug(const char *str) {
	TBuf<100> buf;
	while (*str != '\0')
		buf.Append(*str++);
	RFileLogger::WriteFormat(_L("debug"), _L("bantumi.txt"), EFileLoggingModeAppend, _L("%S"), &buf);
}

CBantumiAppView* CBantumiAppView::NewL(const TRect& aRect, TBool aFullscreen) {
	CBantumiAppView* self = CBantumiAppView::NewLC(aRect, aFullscreen);
	CleanupStack::Pop(self);
	return self;
}

CBantumiAppView* CBantumiAppView::NewLC(const TRect& aRect, TBool aFullscreen) {
	CBantumiAppView* self = new(ELeave) CBantumiAppView();
	CleanupStack::PushL(self);
	self->ConstructL(aRect, aFullscreen);
	return self;
}

void CBantumiAppView::ConstructL(const TRect& aRect, TBool aFullscreen) {
	iDSAStarted = EFalse;
	iEGLInited = EFalse;

	iFrameTickLength = 50;
	iFrameTick = new(ELeave) TInt[iFrameTickLength];
	for (int i = 0; i < iFrameTickLength; i++)
		iFrameTick[i] = 0;

	CreateWindowL();
	iOrigRect = aRect;
	if (aFullscreen) {
		SetExtentToWholeScreen();
		iWinMode = EFullscreen;
	} else {
		SetRect(aRect);
		iWinMode = EWindowed;
	}
	ActivateL();

//	iDrawMode = EEGLWindow;
//	iDrawMode = EDoubleBuf;
	iDrawMode = EDSA;
	SetupBitmap();


	if (HAL::Get(HALData::ESystemTickPeriod, iPeriod) != KErrNone)
		iPeriod = 1000/64;
	else
		iPeriod /= 1000;


	srand(time(NULL));
	iBantumiGL = new BantumiGL(Rect().Width(), Rect().Height());
	if (iDrawMode != EEGLWindow)
		iGLTextFont = makeSymbianGLFont(iGc, Rect().Height());
	else
		iGLTextFont = makeGLFont((const char*) LatinBold13(), 0, 128, 128, ' ', 'z', 'z'+1, ''-1);
	iGLNumFont = makeGLFont((const char*) NumberPlain5(), 0, 32, 32, '0', '9');
	iBantumiGL->setNumFont(iGLNumFont);
	iBantumiGL->setTextFont(iGLTextFont);
	iBantumi = new Bantumi(iBantumiGL);
	iBantumiGL->useScissor(iDrawMode != EEGLWindow);

	iPeriodic = CPeriodic::NewL(CActive::EPriorityIdle);
	iStarted = EFalse;
	SetForeground(ETrue);

	iFrame = 0;
	iLastFPSFrame = 0;
	iLastFPSTime = (TUint)-1;
	iTime = 0;
	iLastFPS = 0;
	iLastResetTick = 0;
	iLastAfterTick = 0;

}

void CBantumiAppView::InitEGL(TBool aPixmap) {
	EGLConfig config;
	TBool reinit = EFalse;

	if (!iEGLInited) {
		iDisplay = eglGetDisplay(EGL_DEFAULT_DISPLAY);
		if (!iDisplay) {
			_LIT(KGetDisplayFailed, "eglGetDisplay failed");
			User::Panic(KGetDisplayFailed, eglGetError());
		}

		if (eglInitialize(iDisplay, NULL, NULL) == EGL_FALSE) {
			_LIT(KInitializeFailed, "eglInitialize failed");
			User::Panic(KInitializeFailed, eglGetError());
		}

		iEGLInited = ETrue;
	} else {
		eglMakeCurrent(iDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
		eglDestroySurface(iDisplay, iSurface);
		eglDestroyContext(iDisplay, iContext);
		reinit = ETrue;
	}

	EGLConfig *configList = NULL;
	EGLint numConfigs = 0, configSize = 0;

	if (eglGetConfigs(iDisplay, configList, configSize, &numConfigs) == EGL_FALSE) {
		_LIT(KGetConfigsFailed, "eglGetConfigs failed");
		User::Panic(KGetConfigsFailed, eglGetError());
	}

	configSize = numConfigs;
	configList = (EGLConfig*) User::Alloc(sizeof(EGLConfig)*configSize);
	if (!configList) {
		_LIT(KConfigAllocFailed, "config alloc failed");
		User::Panic(KConfigAllocFailed, eglGetError());
	}

	TDisplayMode mode = Window().DisplayMode();
	TInt bufferSize = 0;

	switch (mode) {
		case EColor4K: bufferSize = 12; break;
		case EColor64K: bufferSize = 16; break;
		case EColor16M: bufferSize = 24; break;
		case EColor16MU: bufferSize = 32; break;
		default:
			_LIT(KModeError, "unsupported displaymode");
			User::Panic(KModeError, 0);
			break;
	}

	TInt type = EGL_WINDOW_BIT;
	if (aPixmap)
		type = EGL_PIXMAP_BIT;
	EGLint attribList[] = {
		EGL_BUFFER_SIZE, bufferSize, EGL_DEPTH_SIZE, 16,
		EGL_SURFACE_TYPE, type,
		EGL_NONE
	};

	if (eglChooseConfig(iDisplay, attribList, configList, configSize, &numConfigs) == EGL_FALSE) {
		_LIT(KChooseConfigFailed, "eglChooseConfig failed");
		User::Panic(KChooseConfigFailed, eglGetError());
	}

	int n;
	for (n = 0; n < numConfigs; n++) {
		EGLint red, green, blue, size;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_RED_SIZE, &red) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_GREEN_SIZE, &green) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_BLUE_SIZE, &blue) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_BUFFER_SIZE, &size) != EGL_TRUE) continue;

		if (size > bufferSize) continue;
		if (red + green + blue < bufferSize) continue;
		break;
	}
	if (n == numConfigs) {
		_LIT(KNoSuitableConfigs, "No suitable configs");
		User::Panic(KNoSuitableConfigs, 0);
	}

	config = configList[n];
	User::Free(configList);

	iContext = eglCreateContext(iDisplay, config, NULL, NULL);
	if (!iContext) {
		_LIT(KCreateContextFailed, "eglCreateContext failed");
		User::Panic(KCreateContextFailed, eglGetError());
	}

	if (aPixmap)
		iSurface = eglCreatePixmapSurface(iDisplay, config, iBitmap, NULL);
	else
		iSurface = eglCreateWindowSurface(iDisplay, config, &Window(), NULL);

	if (!iSurface) {
		_LIT(KCreateWindowSurfaceFailed, "eglCreateWindowSurface failed");
		_LIT(KCreatePixmapSurfaceFailed, "eglCreatePixmapSurface failed");
		if (aPixmap)
			User::Panic(KCreatePixmapSurfaceFailed, eglGetError());
		else
			User::Panic(KCreateWindowSurfaceFailed, eglGetError());
	}

	if (eglMakeCurrent(iDisplay, iSurface, iSurface, iContext) == EGL_FALSE) {
		_LIT(KMakeCurrentFailed, "eglMakeCurrent failed");
		User::Panic(KMakeCurrentFailed, eglGetError());
	}

	if (reinit && iBantumiGL)
		iBantumiGL->reinitGL(Rect().Width(), Rect().Height());
}

void CBantumiAppView::SetForeground(TBool aForeground) {
	if (aForeground) {
		if (!iStarted) {
			iPeriodic->Start(100, 100, TCallBack(CBantumiAppView::StaticCallback, this));
			iLastTick = User::TickCount();
		}
	} else {
		if (iStarted)
			iPeriodic->Cancel();
	}
	iStarted = aForeground;
	if (iBantumi)
		iBantumi->setPaused(!aForeground);
}

void CBantumiAppView::SetDrawMode(TDrawMode aMode, TBool aReset) {
	if (aMode == iDrawMode && !aReset)
		return;

	delete iDrawer;
	iDrawer = NULL;

	TDrawMode lastMode = iDrawMode;
	iDrawMode = aMode;

	TInt err;
	switch (iDrawMode) {
	case EEGLWindow:
		InitEGL(EFalse);
		break;

	case EDoubleBuf:
		if (lastMode != EDSA)
			InitEGL(ETrue);
		break;

	case EDSA:
		if (lastMode != EDoubleBuf)
			InitEGL(ETrue);
		iDrawer = CDirectScreenAccess::NewL(iEikonEnv->WsSession(), *(iEikonEnv->ScreenDevice()), Window(), *this);
		iEikonEnv->WsSession().Flush();
		TRAP(err, iDrawer->StartL());
		if (err == KErrNone) {
			iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
			iDSAStarted = ETrue;
		}
		break;
	}


}

void CBantumiAppView::SetWindowMode(TWindowMode aMode) {
	if (aMode == iWinMode)
		return;

	iWinMode = aMode;

	if (iWinMode == EFullscreen)
		SetExtentToWholeScreen();
	else
		SetRect(iOrigRect);

	SetupBitmap();

}

void CBantumiAppView::SetupBitmap() {
	if (iFont && iGc)
		iGc->DiscardFont();
	delete iBitmap;
	delete iDevice;
	delete iGc;

	iBitmap = new(ELeave) CWsBitmap(iEikonEnv->WsSession());
	iBitmap->Create(Rect().Size(), iEikonEnv->ScreenDevice()->DisplayMode());

	iDevice = CFbsBitmapDevice::NewL(iBitmap);
	TInt err = iDevice->CreateContext(iGc);
	if (err != KErrNone) {
		_LIT(KPanicMsg, "CFbsBitmapDevice::CreateContext");
		User::Panic(KPanicMsg, err);
	}

//	iFont = LatinBold13();
	iFont = LatinPlain12();
//	iFont = NumberPlain5();

	SetDrawMode(iDrawMode, ETrue);
}

CBantumiAppView::CBantumiAppView() {
}


CBantumiAppView::~CBantumiAppView() {
	if (iFont && iGc)
		iGc->DiscardFont();
	delete iBitmap;
	delete iDevice;
	delete iGc;
	delete iDrawer;
	delete iPeriodic;
	delete iBantumi;
	delete iBantumiGL;
	delete iGLNumFont;
	delete iGLTextFont;

	eglMakeCurrent(iDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
	eglDestroySurface(iDisplay, iSurface);
	eglDestroyContext(iDisplay, iContext);
	eglTerminate(iDisplay);
}


void CBantumiAppView::Draw(const TRect& aRect) const {
	CWindowGc& gc = SystemGc();
//	TRect rect = Rect();
//	gc.Clear(rect);
	if (iDrawMode == EEGLWindow) {
		eglCopyBuffers(iDisplay, iSurface, iBitmap);
		eglWaitGL();
	}
	gc.BitBlt(TPoint(0, 0), iBitmap);
}

void CBantumiAppView::Update() {
	TInt err = KErrNone;
	if (!iDSAStarted) {
		TRAP(err, iDrawer->StartL());
		if (err == KErrNone) {
			iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
			iDSAStarted = ETrue;
		}
	}
	if (err == KErrNone) {
		iDrawer->Gc()->BitBlt(TPoint(0, 0), iBitmap);
	}
}

void CBantumiAppView::Restart(RDirectScreenAccess::TTerminationReasons aReason) {
/*
	TRAPD(err, iDrawer->StartL());
	if (err == KErrNone) {
		iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
		iDSAStarted = ETrue;
	}
*/
}

void CBantumiAppView::AbortNow(RDirectScreenAccess::TTerminationReasons aReason) {
	SetForeground(EFalse);
	iDSAStarted = EFalse;
}

TInt CBantumiAppView::StaticCallback(TAny* aArg) {
	CBantumiAppView* view = (CBantumiAppView*) aArg;
	view->Callback();
	return 0;
}

void CBantumiAppView::Callback() {
	iFrame++;

	if (iBantumiGL->exit()) {
		static_cast<CBantumiAppUi*>(iEikonEnv->AppUi())->DoExit();
	}

	iBantumi->update();
	iBantumiGL->draw(iTime);

	TInt tick = User::TickCount();
	iRelTick += tick - iLastTick;
	iTime = iRelTick*iPeriod;
	iLastTick = tick;

/*	if (iDrawMode == EDoubleBuf || iDrawMode == EDSA) {
		TPoint point(2, 2 + iFont->HeightInPixels());
		TBuf<20> buf;
		buf.Format(_L("%d"), iLastFPS);
		iGc->UseFont(iFont);
		iGc->SetPenColor(KRgbWhite);
		iGc->SetDrawMode(CGraphicsContext::EDrawModeXOR);
		iGc->DrawText(buf, point);
		TInt ticklength = iRelTick - iFrameTick[iFrame % iFrameTickLength];
		buf.Format(_L("%d"), ticklength);
		point.iX = iBitmap->SizeInPixels().iWidth - 2 - iFont->TextWidthInPixels(buf);
		iGc->DrawText(buf, point);
	}
*/
	iFrameTick[iFrame % iFrameTickLength] = iRelTick;

//	eglCopyBuffers(iDisplay, iSurface, iBitmap);
	if (iDrawMode == EEGLWindow)
		eglSwapBuffers(iDisplay, iSurface);
	else if (iDrawMode == EDoubleBuf) {
		eglWaitGL();
		DrawNow();
	} else if (iDrawMode == EDSA) {
		eglWaitGL();
		Update();
	}


	if (iLastFPSTime == (TUint)-1)
		iLastFPSTime = iTime;

	if ((tick - iLastResetTick)*iPeriod >= 3000) {
		User::ResetInactivityTime();
		iLastResetTick = tick;
	}
	if ((tick - iLastAfterTick)*iPeriod >= 1500) {
		User::After(0);
		iLastAfterTick = tick;
	}

	if (iTime - iLastFPSTime > 1000) {
		iLastFPS = 1000*(iFrame - iLastFPSFrame)/(iTime - iLastFPSTime);
		RFileLogger::WriteFormat(_L("debug"), _L("bantumi.txt"), EFileLoggingModeAppend, _L("%d fps"), iLastFPS);
		iLastFPSTime = iTime;
		iLastFPSFrame = iFrame;
	}
}

TKeyResponse CBantumiAppView::OfferKeyEventL(const TKeyEvent& aKeyEvent, TEventCode aType) {
	if (aType != EEventKeyDown)
		return EKeyWasNotConsumed;

	if (iBantumi) {
		switch (aKeyEvent.iScanCode) {
		case EStdKeyDevice3:
			iBantumi->pressed(SELECT);
			break;
		case EStdKeyUpArrow:
			iBantumi->pressed(UP);
			break;
		case EStdKeyDownArrow:
			iBantumi->pressed(DOWN);
			break;
		case EStdKeyLeftArrow:
			iBantumi->pressed(LEFT);
			break;
		case EStdKeyRightArrow:
			iBantumi->pressed(RIGHT);
			break;
		default:
			return EKeyWasNotConsumed;
		}
		return EKeyWasConsumed;
	}
	return EKeyWasNotConsumed;
}

